/***************************************************************************
 *
 * gpm-prefs.c : GNOME Power Preferences
 *
 * Copyright (C) 2005 Richard Hughes, <richard@hughsie.com>
 *
 * Taken in part from:
 * - GNOME Tutorial, example 1
 * http://www.gnome.org/~newren/tutorials/developing-with-gnome/html/apd.html
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 **************************************************************************/

#include <glib.h>
#include <math.h>
#include <dbus/dbus-glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gconf/gconf-client.h>
#if defined(HAVE_LIBNOTIFY)
#include <libnotify/notify.h>
#endif

#include "gpm-common.h"
#include "gpm-prefs.h"
#include "gpm-main.h"
#include "hal-glib.h"

static GladeXML *all_pref_widgets;
static gboolean isVerbose;

/** Convenience function to call libnotify
 *
 *  @param  content		The content text, e.g. "Battery low"
 *  @param  value		The urgency, e.g NOTIFY_URGENCY_CRITICAL
 */
static void
use_libnotify (const char *content, const int urgency)
{
#if defined(HAVE_LIBNOTIFY)
	NotifyIcon *icon = notify_icon_new_from_uri (GPM_DATA "gnome-power.png");
	const char *summary = NICENAME;
	NotifyHandle *n = notify_send_notification (NULL, /* replaces nothing 	*/
			   NULL,
			   urgency,
			   summary, content,
			   icon, /* no icon 			*/
			   TRUE, NOTIFY_TIMEOUT,
			   NULL,
			   NULL, /* no user data 		*/
			   0);   /* no actions 			*/
	notify_icon_destroy(icon);
	if (!n)
		g_warning ("failed to send notification (%s)", content);
#else
	GtkWidget *widget;
	widget = gnome_message_box_new (content, 
                                GNOME_MESSAGE_BOX_WARNING,
                                GNOME_STOCK_BUTTON_OK, 
                                NULL);
	gtk_window_set_title (GTK_WINDOW (widget), NICENAME);
	gtk_widget_show (widget);
#endif
}

/** queries org.gnome.GnomePowerManager.isOnBattery
 *
 *  @param  value	return value, passed by ref
 *  @return		TRUE for success, FALSE for failure
 */
gboolean
gpm_is_on_mains (gboolean *value)
{
	DBusGConnection *session_connection;
	DBusGProxy *gpm_proxy;
	GError *error = NULL;
	gboolean retval;

	session_connection = get_session_connection ();
	gpm_proxy = dbus_g_proxy_new_for_name (session_connection,
			GPM_DBUS_SERVICE, GPM_DBUS_PATH, GPM_DBUS_INTERFACE);
	retval = TRUE;
	if (!dbus_g_proxy_call (gpm_proxy, "isOnBattery", &error, 
			G_TYPE_INVALID,
			G_TYPE_BOOLEAN, value, G_TYPE_INVALID)) {
		dbus_glib_error (error);
		*value = FALSE;
		retval = FALSE;
	}
	g_object_unref (G_OBJECT (gpm_proxy));
	return retval;
}

/** Sets/Hides GTK visibility
 *
 *  @param  widgetname		the libglade widget name
 *  @param  set			should widget be visible?
 */
static void
gtk_set_visibility (const char *widgetname, gboolean set)
{
	GtkWidget *widget;

	g_return_if_fail (widgetname);

	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	if (!widget) {
		g_warning ("gtk_set_visibility: widget '%s' not found", widgetname);
		return;
	}

	if (set)
		gtk_widget_show_all (widget);
	else
		gtk_widget_hide_all (widget);
}

/** Sets/Clears GTK Checkbox
 *
 *  @param  widgetname		the libglade widget name
 *  @param  set			should check be ticked?
 */
static void
gtk_set_check (const char *widgetname, gboolean set)
{
	GtkWidget *widget;

	g_return_if_fail (widgetname);

	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	if (!widget) {
		g_warning ("widget '%s' failed to load, aborting", widgetname);
		return;
	}
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), set);
}

/** Shows/hides/renames controls based on hasData, i.e. what hardware is in the system.
 *
 */
static void
recalc (void)
{
	GtkWidget *widget;
	GConfClient *client = gconf_client_get_default ();

	gboolean hasBatteries;
	gboolean hasAcAdapter;
	gboolean hasButtonPower;
	gboolean hasButtonSleep;
	gboolean hasButtonLid;
	gboolean hasLCD;
	gboolean hasDisplays;
	gboolean useGnomeScreensaver;

	/* checkboxes */
	gboolean displayIcon = gconf_client_get_bool (client, GCONF_ROOT "general/display_icon", NULL);
	gboolean displayIconFull = gconf_client_get_bool (client, GCONF_ROOT "general/display_icon_full", NULL);
	gtk_set_check ("checkbutton_display_icon", displayIcon);
	gtk_set_check ("checkbutton_display_icon_full", displayIconFull);

	hasBatteries =   (hal_num_devices_of_capability ("battery") > 0);
	hasAcAdapter =   (hal_num_devices_of_capability ("ac_adapter") > 0);
	hasButtonPower = (hal_num_devices_of_capability_with_value ("button", "button.type", "power") > 0);
	hasButtonSleep = (hal_num_devices_of_capability_with_value ("button", "button.type", "sleep") > 0);
	hasButtonLid =   (hal_num_devices_of_capability_with_value ("button", "button.type", "lid") > 0);
#if 0
	hasHardDrive =   (hal_num_devices_of_capability_with_value ("storage", "storage.bus", "ide") > 0);
#endif
	hasLCD = 	 (hal_num_devices_of_capability ("laptop_panel") > 0);

#if defined(HAVE_GSCREENSAVER)
	hasDisplays = gconf_client_get_bool (client, "/apps/gnome-screensaver/dpms_enabled", NULL);
	useGnomeScreensaver = TRUE;
#else
	hasDisplays = FALSE;
	useGnomeScreensaver = FALSE;
#endif

	/* frame labels */
	if (hasBatteries) {
		widget = glade_xml_get_widget (all_pref_widgets, "label_frame_ac");
		gtk_label_set_markup (GTK_LABEL (widget), "<b>Running on AC adapter</b>");
		widget = glade_xml_get_widget (all_pref_widgets, "label_frame_batteries");
		gtk_label_set_markup (GTK_LABEL (widget), "<b>Running on batteries</b>");
	} else {
		widget = glade_xml_get_widget (all_pref_widgets, "label_frame_ac");
		gtk_label_set_markup (GTK_LABEL (widget), "<b>Configuration</b>");
	}

	/* top frame */
	gtk_set_visibility ("frame_batteries", hasBatteries);
	gtk_set_visibility ("combobox_battery_critical", hasBatteries);
	gtk_set_visibility ("label_battery_critical_action", hasBatteries);
	gtk_set_visibility ("label_battery_critical", hasBatteries);
	gtk_set_visibility ("label_battery_low", hasBatteries);
	gtk_set_visibility ("hscale_battery_low", hasBatteries);
	gtk_set_visibility ("hscale_battery_critical", hasBatteries);
	/* assumes only battery options are in this frame */
	gtk_set_visibility ("frame_other_options", hasBatteries);

	/* options */
	gtk_set_visibility ("combobox_button_lid", hasButtonLid);
	gtk_set_visibility ("label_button_lid", hasButtonLid);

	gtk_set_visibility ("combobox_button_power", hasButtonPower);
	gtk_set_visibility ("label_button_power", hasButtonPower);

	gtk_set_visibility ("combobox_button_suspend", hasButtonSleep);
	gtk_set_visibility ("label_button_suspend", hasButtonSleep);

	gtk_set_visibility ("combobox_ac_fail", hasAcAdapter);
	gtk_set_visibility ("label_ac_fail", hasAcAdapter);

	/* variables */
	gtk_set_visibility ("hscale_ac_brightness", hasLCD);
	gtk_set_visibility ("label_ac_brightness", hasLCD);
	gtk_set_visibility ("hscale_batteries_brightness", hasLCD);
	gtk_set_visibility ("label_batteries_brightness", hasLCD);

	gtk_set_visibility ("hscale_ac_display", hasDisplays);
	gtk_set_visibility ("label_ac_display", hasDisplays);
	gtk_set_visibility ("hscale_batteries_display", hasDisplays & hasBatteries);
	gtk_set_visibility ("label_batteries_display", hasDisplays & hasBatteries);

	/* set the display stuff to set gnome-screensaver dpms timeout */
	gtk_set_visibility ("hscale_ac_display", hasDisplays);
	gtk_set_visibility ("label_ac_display", hasDisplays);
	gtk_set_visibility ("hscale_batteries_display", hasDisplays & hasBatteries);
	gtk_set_visibility ("label_batteries_display", hasDisplays & hasBatteries);
	gtk_set_visibility ("button_gnome_screensave", useGnomeScreensaver);
}

/** Callback for gconf_key_changed
 *
 */
static void
callback_gconf_key_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
	g_return_if_fail (client);
	if (gconf_entry_get_value (entry) == NULL)
		return;

	/* 
	 * just recalculate the UI, as the gconf keys are read there.
	 * this removes the need for lots of global variables.
	 */
#if 0
	recalc ();
#endif
}

/** Callback for combo_changed
 *
 */
static void
callback_combo_changed (GtkWidget *widget, gpointer user_data)
{
	GConfClient *client;
	gint value;
	gchar *policypath;
	gchar *policyoption;

	g_return_if_fail (widget);
	client = gconf_client_get_default ();
	value = gtk_combo_box_get_active(GTK_COMBO_BOX (widget));
	policypath = g_object_get_data ((GObject*) widget, "policypath");
	g_return_if_fail (policypath);

	g_debug ("[%s] = (%i)", policypath, value);

	policyoption = convert_policy_to_string (value);
	gconf_client_set_string (client, policypath, policyoption, NULL);
}

/** Callback for hscale_changed
 *
 */
static void
callback_hscale_changed (GtkWidget *widget, gpointer user_data)
{
	GConfClient *client;
	const char *widgetname;
	gchar *policypath;
	gint value;
	gint oldgconfvalue;
	gdouble divisions = -1;
	gboolean onbattery;

	g_return_if_fail (widget);

	client = gconf_client_get_default ();
	policypath = g_object_get_data ((GObject*) widget, "policypath");

	value = (int) gtk_range_get_value (GTK_RANGE (widget));
	oldgconfvalue = gconf_client_get_int (client, policypath, NULL);

	/* 
	 * Code for divisions of 10 seconds, unfinished
	 *
	 */
	widgetname = gtk_widget_get_name (widget);
	if (strcmp (widgetname, "hscale_ac_brightness") == 0)
		divisions = 100.0 / hal_get_brightness_steps ();
	else if (strcmp (widgetname, "hscale_batteries_brightness") == 0)
		divisions = 100.0 / hal_get_brightness_steps ();
	else if (strcmp (widgetname, "hscale_ac_computer") == 0)
		divisions = 5;
	else if (strcmp (widgetname, "hscale_batteries_computer") == 0)
		divisions = 5;

	if (divisions > 0) {
		double double_segment = ((gdouble) value / divisions);
		double v2 = ceil(double_segment) * divisions;
		gtk_range_set_value (GTK_RANGE (widget), v2);
		/* if not different, then we'll return, and wait for the proper event */
		if ((int) v2 != (int) value)
			return;
	}

	/*
	 * if calculated value not different to existing gconf value, 
	 * then no point continuing
	 */
	if (oldgconfvalue == value)
		return;

	/* if this is hscale for battery_low, then set upper range of hscale for 
	 * battery_critical maximum to value
	 * (This stops criticalThreshold > lowThreshold)
	 */
	if (strcmp (widgetname, "hscale_battery_low") == 0) {
		GtkWidget *widget2;
		widget2 = glade_xml_get_widget (all_pref_widgets, "hscale_battery_critical");
		gtk_range_set_range (GTK_RANGE (widget2), 0, value);
	}

	/* for AC and battery, change the brightness in real-time */
	if (gpm_is_on_mains (&onbattery)) {
		if ((!onbattery && strcmp (widgetname, "hscale_ac_brightness") == 0) || 
		    (onbattery && strcmp (widgetname, "hscale_batteries_brightness") == 0))
			hal_set_brightness (value);
	} else
		g_warning (GPM_DBUS_SERVICE ".isOnBattery failed");

	g_return_if_fail (policypath);
	g_debug ("[%s] = (%i)", policypath, value);
	gconf_client_set_int (client, policypath, value, NULL);
}

/** Callback for button_help
 *
 */
static void
callback_help (GtkWidget *widget, gpointer user_data)
{
	/* for now, show website */
	gnome_url_show (GPMURL, NULL);
}

/** Callback for check_changed
 *
 */
static void
callback_check_changed (GtkWidget *widget, gpointer user_data)
{
	GConfClient *client;
	gboolean value;
	gchar *policypath;

	g_return_if_fail (widget);

	client = gconf_client_get_default ();
	value = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
	policypath = g_object_get_data ((GObject*) widget, "policypath");
	g_return_if_fail (policypath);
	/*int policytype = (int) g_object_get_data ((GObject*) widget, "policytype");*/

	g_debug ("[%s] = (%i)", policypath, value);
	gconf_client_set_bool (client, policypath, value, NULL);
}

/** Prints program usage.
 *
 */
static void
print_usage (void)
{
	g_print ("\nusage : gnome-power-preferences [--verbose] [--help]\n");
	g_print (
		"\n"
		"        --verbose               Show extra debugging\n"
		"        --help                  Show this information and exit\n"
		"\n");
}

/** simple callback formatting a GtkScale to "10.0%"
 *
 */
static gchar*
format_value_callback_percent (GtkScale *scale, gdouble value)
{
	return g_strdup_printf ("%4.1f%%", value);
}

/** simple callback that converts minutes into pretty text
 *
 */
static gchar*
format_value_callback_time (GtkScale *scale, gdouble value)
{
	gchar unitstring[32];
	GString *strvalue;

	if ((gint) value == 0)
		return g_strdup_printf ("Never");

	strvalue = get_timestring_from_minutes (value);
	strcpy (unitstring, strvalue->str);
	g_string_free (strvalue, TRUE);
	return g_strdup_printf ("%s", unitstring);
}

/** Sets the comboboxes up to the gconf value, and sets up callbacks.
 *
 *  @param  widgetname		the libglade widget name
 *  @param  policypath		the GConf policy path, 
 *				e.g. "policy/ac/brightness"
 *  @param  policytype		the policy type, e.g. POLICY_PERCENT
 */
static void
combo_setup_action (const char *widgetname, const char *policypath, int policytype)
{
	GConfClient *client;
	GtkWidget *widget;
	gchar *policyoption;
	gint value;
	g_return_if_fail (widgetname);

	client = gconf_client_get_default ();
	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	g_return_if_fail (widget);

	g_object_set_data ((GObject*) widget, "policypath", (gpointer) policypath);
	g_object_set_data ((GObject*) widget, "policytype", (gpointer) policytype);

	if (policytype == POLICY_CHOICE) {
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Do nothing"));
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Send warning"));
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Suspend"));
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Hibernate"));
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Shutdown"));
	} else if (policytype == POLICY_SLEEP) {
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Suspend"));
		gtk_combo_box_append_text (GTK_COMBO_BOX (widget), _("Hibernate"));
	}

	policyoption = gconf_client_get_string (client, policypath, NULL);

	if (!policyoption) {
		g_warning ("gconf_client_get_string for widget '%s' failed (policy='%s')!!", widgetname, policypath);
		return;
	}

	value = convert_string_to_policy (policyoption);
	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), value);
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (callback_combo_changed), NULL);
}

/** Sets the hscales up to the gconf value, and sets up callbacks.
 *
 *  @param  widgetname		the libglade widget name
 *  @param  policypath		the GConf policy path, 
 *				e.g. "policy/ac/brightness"
 *  @param  policytype		the policy ptye, e.g. POLICY_PERCENT
 */
static void
hscale_setup_action (const char *widgetname, const char *policypath, int policytype)
{
	GConfClient *client;
	GtkWidget *widget;
	gint value;

	g_return_if_fail (widgetname);
	g_return_if_fail (policypath);

	client = gconf_client_get_default ();
	widget = glade_xml_get_widget (all_pref_widgets, widgetname);

	g_object_set_data ((GObject*) widget, "policypath", (gpointer) policypath);
	g_object_set_data ((GObject*) widget, "policytype", (gpointer) policytype);

	value = gconf_client_get_int (client, policypath, NULL);
	g_debug ("'%s' -> [%s] = (%i)", widgetname, policypath, value);

	if (policytype == POLICY_PERCENT)
		g_signal_connect (G_OBJECT (widget), "format-value", G_CALLBACK (format_value_callback_percent), NULL);
	else
		g_signal_connect (G_OBJECT (widget), "format-value", G_CALLBACK (format_value_callback_time), NULL);
	gtk_range_set_value (GTK_RANGE (widget), (int) value);
	g_signal_connect (G_OBJECT (widget), "value-changed", G_CALLBACK (callback_hscale_changed), NULL);
}

/** Sets the checkboxes up to the gconf value, and sets up callbacks.
 *
 *  @param  widgetname		the libglade widget name
 *  @param  policypath		the GConf policy path, 
 *				e.g. "policy/ac/brightness"
 */
static void
checkbox_setup_action (const char *widgetname, const char *policypath)
{
	GConfClient *client;
	GtkWidget *widget;
	gboolean value;

	g_return_if_fail (widgetname);
	g_return_if_fail (policypath);

	client = gconf_client_get_default ();
	widget = glade_xml_get_widget (all_pref_widgets, widgetname);
	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (callback_check_changed), NULL);
	g_object_set_data ((GObject*) widget, "policypath", (gpointer) policypath);

	value = gconf_client_get_bool (client, policypath, NULL);
	g_debug ("'%s' -> [%s] = (%i)", widgetname, policypath, value);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), value);
}

/** Main program
 *
 */
int
main (int argc, char **argv)
{
	GtkWidget *widget = NULL;
	gint a;
	gint value;
	GConfClient *client;

	gtk_init (&argc, &argv);
	gconf_init (argc, argv, NULL);

	isVerbose = FALSE;
	for (a=1; a < argc; a++) {
		if (strcmp (argv[a], "--verbose") == 0)
			isVerbose = TRUE;
		else if (strcmp (argv[a], "--help") == 0) {
			print_usage ();
			return EXIT_SUCCESS;
		}
	}

	if (!isVerbose)
		g_log_set_handler (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, g_log_ignore, NULL);

	/* load the interface */
	all_pref_widgets = glade_xml_new (GPM_DATA "preferences.glade", NULL, NULL);
	if (!all_pref_widgets)
		g_error ("glade file failed to load, aborting");

	/* Get the GconfClient, tell it we want to monitor /apps/gnome-power */
	client = gconf_client_get_default ();
	gconf_client_add_dir (client, GCONF_ROOT_SANS_SLASH, GCONF_CLIENT_PRELOAD_NONE, NULL);
	gconf_client_notify_add (client, GCONF_ROOT_SANS_SLASH, callback_gconf_key_changed, widget, NULL, NULL);
#if defined(HAVE_GSCREENSAVER)
	gconf_client_notify_add (client, "/apps/gnome-screensaver", callback_gconf_key_changed, widget, NULL, NULL);
#endif
	/* Get the main_window quit */
	widget = glade_xml_get_widget (all_pref_widgets, "window_preferences");
	if (!widget)
		g_error ("Main window failed to load, aborting");
	g_signal_connect (G_OBJECT (widget), "delete_event", G_CALLBACK (gtk_main_quit), NULL);

#if defined(HAVE_LIBNOTIFY)
	/* initialise libnotify */
	if (!notify_glib_init(NICENAME, NULL))
		g_error ("Cannot initialise libnotify!");
#endif

	/* Get the help and quit buttons */
	widget = glade_xml_get_widget (all_pref_widgets, "button_close");
	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (gtk_main_quit), NULL);
	widget = glade_xml_get_widget (all_pref_widgets, "button_help");
	g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (callback_help), NULL);

	/* set gtk enables/disables */
	recalc ();

	/* disable these until the backend code is in place */
	gtk_set_visibility ("checkbutton_hdd_spindown", FALSE);

	/* checkboxes */
	checkbox_setup_action ("checkbutton_display_icon",
		GCONF_ROOT "general/display_icon");
	checkbox_setup_action ("checkbutton_display_icon_full",
		GCONF_ROOT "general/display_icon_full");

	/* comboboxes */
	combo_setup_action ("combobox_button_power",
		GCONF_ROOT "policy/button_power", POLICY_CHOICE);
	combo_setup_action ("combobox_button_suspend",
		GCONF_ROOT "policy/button_suspend", POLICY_CHOICE);
	combo_setup_action ("combobox_button_lid",
		GCONF_ROOT "policy/button_lid", POLICY_CHOICE);
	combo_setup_action ("combobox_ac_fail",
		GCONF_ROOT "policy/ac_fail", POLICY_CHOICE);
	combo_setup_action ("combobox_battery_critical",
		GCONF_ROOT "policy/battery_critical", POLICY_CHOICE);
	combo_setup_action ("combobox_sleep_type",
		GCONF_ROOT "policy/sleep_type", POLICY_SLEEP);

	/* sliders */
	hscale_setup_action ("hscale_ac_computer", 
		GCONF_ROOT "policy/ac/sleep_computer", POLICY_TIME);
	hscale_setup_action ("hscale_ac_display", 
		GCONF_ROOT "policy/ac/sleep_display", POLICY_TIME);
	hscale_setup_action ("hscale_ac_brightness", 
		GCONF_ROOT "policy/ac/brightness", POLICY_PERCENT);
	hscale_setup_action ("hscale_batteries_computer", 
		GCONF_ROOT "policy/battery/sleep_computer", POLICY_TIME);
	hscale_setup_action ("hscale_batteries_display", 
		GCONF_ROOT "policy/battery/sleep_display", POLICY_TIME);
	hscale_setup_action ("hscale_batteries_brightness", 
		GCONF_ROOT "policy/battery/brightness", POLICY_PERCENT);
	hscale_setup_action ("hscale_battery_low", 
		GCONF_ROOT "general/threshold_low", POLICY_PERCENT);
	hscale_setup_action ("hscale_battery_critical", 
		GCONF_ROOT "general/threshold_critical", POLICY_PERCENT);

	/* set up upper limit for battery_critical */
	widget = glade_xml_get_widget (all_pref_widgets, "hscale_battery_low");
	value = (gint) gtk_range_get_value (GTK_RANGE (widget));
	widget = glade_xml_get_widget (all_pref_widgets, "hscale_battery_critical");
	gtk_range_set_range (GTK_RANGE (widget), 0, value);

#if defined(HAVE_GSCREENSAVER)
	gboolean hasDisplays = gconf_client_get_bool (client, "/apps/gnome-screensaver/dpms_enabled", NULL);
	if (!hasDisplays)
		use_libnotify ("You have not got DPMS support enabled in gnome-screensaver. You cannot cannot change the screen shutdown time using this program.", NOTIFY_URGENCY_NORMAL);
#endif
	gtk_main ();
	return 0;
}
